# -*- coding: utf-8 -*-

import dataiku
from dataiku.core import flow
from dataiku.customrecipe import *
import shutil
import os
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,  # avoid getting log from 3rd party module
                    format='RuleGeneration plugin %(levelname)s - %(message)s')


##########################################################################
# COPY THE PLUGIN PYTHON'S LIB TO THE DSS LIB
##########################################################################

# In the Scoring phase, dss won't have access to the plugin's python lib,
# so we need to copy it to the instance's lib
root_path = dataiku.get_custom_variables()['dip.home']
plugin_path = os.path.join(
    root_path, "plugins/installed/rules-generation/python-lib/RuleGenerationLib")
dss_lib_path = os.path.join(root_path, "lib/python/RuleGenerationLib/")

if os.path.isdir(dss_lib_path):
    logger.info('Delete and recreate the package inside dss lib folder')
    shutil.rmtree(dss_lib_path)
    os.makedirs(dss_lib_path)
else:
    logger.info('Copy package to dss lib folder')
    os.makedirs(dss_lib_path)


for item in os.listdir(plugin_path):
    s = os.path.join(plugin_path, item)
    d = os.path.join(dss_lib_path, item)
    if os.path.isdir(s):
        shutil.copytree(s, d, symlinks=False, ignore=None)
    else:
        shutil.copy2(s, d)


from RuleGenerationLib import MltaskDeployment


##########################################################################
# LOADING ALL REQUIRED INFO AND
# SETTING VARIABLES
##########################################################################

# Recipe inputs
input_ = get_input_names_for_role('TrainSet')
trainset_name = input_[0].split('.')[1]  # TO DO: refactor this quick fix

# General parameters
model_name = get_recipe_config()['model_name']
ml_type = get_recipe_config()['ml_type']

try:
    target = get_recipe_config()['Target']
except:
    raise ValuError("You need to define the target column!")

optimisation_metric = get_recipe_config()['optimisation_metric']

# Algorithm parameters
max_depth = get_recipe_config()['max_depth']

alpha = get_recipe_config()['alpha']

# Optimiszation constrain parameters
max_support_threshold = 1  # get_recipe_config()['max_support_threshold']

performance_drop_threshold = get_recipe_config()['performance_drop_threshold']

# Advanced parameters
ignored_col_list = get_recipe_config()['ignored_col_list']
min_samples_leaf = get_recipe_config()['min_samples_leaf']
min_samples_leaf = min_samples_leaf / 100.


# Sanity check
assert model_name is not None, "You need to define a model name!"
assert target is not None, "You need to define the target column!"
assert max_depth is not None, "You need to define the max number of clauses in a rule!"
assert alpha is not None, "You need to define the regularisation parameter!"
assert performance_drop_threshold is not None, "You need to define a performance drop threshold!"

REGRESSION_METRICS = ['r2', 'neg_mean_absolute_error',
                      'neg_mean_squared_error', 'explained_variance']
CLASSIF_METRICS = ['f1', 'roc_auc', 'accuracy', 'precision', 'recall']

if ml_type == 'reg' and optimisation_metric not in REGRESSION_METRICS:
    raise ValueError('{0} is not valid. Possible regession metrics are: {1}'.format(
        optimisation_metric, REGRESSION_METRICS))

if ml_type == 'clf' and optimisation_metric not in CLASSIF_METRICS:
    raise ValueError('{0} is not valid. Possible classification metrtics are: {1}'.format(
        optimisation_metric, CLASSIF_METRICS))


# TODO: let user configure this ?
n_trees = 20  # get_recipe_config()['n_trees_list']


client = dataiku.api_client()
project = client.get_project(dataiku.get_custom_variables()['projectKey'])

DSS_METRICS_REMAP = {
    'r2': 'R2',
    'neg_mean_absolute_error': 'MAE',
    'neg_mean_squared_error': 'MSE',
    'explained_variance': 'EVS',
    'f1': 'F1',
    'roc_auc': 'ROC_AUC',
    'accuracy': 'ACCURACY',
    'precision': 'PRECISION',
    'recall': 'RECALL',
    'neg_log_loss': 'LOG_LOSS',
}


##########################################################################
# APPLY THE ALGORITHM
##########################################################################

# create the first mltask to retrieve the columns's name AFTER the
# preprocessing step


raw_ml_task = project.create_prediction_ml_task(
    input_dataset=trainset_name, target_variable=target)

raw_ml_task.wait_guess_complete()

ml_task = MltaskDeployment.configure_mltask(ml_task=raw_ml_task,
                                            ignored_column_type_list=[
                                                'TEXT', 'VECTOR'],
                                            algorithm_name='random_forest_{}'.format(
                                                'regression' if ml_type == 'reg' else 'classification'),
                                            algorithm_setting_dict={
                                                'max_tree_depth': [1], 'n_estimators': [1]},
                                            metric=DSS_METRICS_REMAP[
                                                optimisation_metric],
                                            ignored_column_list=ignored_col_list)


try:
    dummy_model = MltaskDeployment.deploy_mltask(
        deployed_model_name=model_name, ml_task=ml_task, project=project, train_dataset=trainset_name)
except:
    raise ValueError(
        'Training phase failed. Please go to the Lab section for more details.')


# now we create the main mltask
raw_ml_task = project.create_prediction_ml_task(
    input_dataset=trainset_name, target_variable=target)
raw_ml_task.wait_guess_complete()


processed_feature_list = dummy_model.get_predictor().features
logger.info("AFTER PREPROCESSING, NUM FEATURE: ", len(processed_feature_list))


custom_algorithm = """from RuleGenerationLib.RuleGeneratorRandomForestParams import RuleGeneratorRandomForestParams
from RuleGenerationLib.RuleGeneratorRandomForest import RuleGeneratorRandomForest
from RuleGenerationLib.RuleGeneratorRegularizerParams import RuleGeneratorRegularizerParams
from RuleGenerationLib.RuleGeneratorRegularizer import RuleGeneratorRegularizer
from RuleGenerationLib.RuleGeneratorParams import RuleGeneratorParams
from RuleGenerationLib.RuleGenerator import RuleGenerator


rule_generator_rf_params = RuleGeneratorRandomForestParams(ml_type = '{0}', optimisation_metric = '{1}', n_trees_list = [{2}], min_samples_leaf_list = [{3}], max_depth_list = [{4}])
rule_generator_rf = RuleGeneratorRandomForest(rule_generator_rf_params)

rule_genrator_regularizer_params = RuleGeneratorRegularizerParams(ml_type = '{0}', alpha = {5})
rule_generator_regularizer = RuleGeneratorRegularizer(rule_genrator_regularizer_params)

rule_generator_params = RuleGeneratorParams(raw_feature_list = {6},
											target = '{7}',
											random_forest_estimator = rule_generator_rf,
											rule_regularizer = rule_generator_regularizer,
											max_support_threshold = {8},
											performance_drop_threshold = {9})

clf = RuleGenerator(rule_generator_params)""".format(ml_type, optimisation_metric, n_trees, min_samples_leaf, max_depth, alpha, processed_feature_list, target, max_support_threshold, performance_drop_threshold)


ml_task = MltaskDeployment.configure_mltask(ml_task=raw_ml_task,
                                            ignored_column_type_list=[
                                                'TEXT', 'VECTOR'],
                                            metric=DSS_METRICS_REMAP[
                                                optimisation_metric],
                                            ignored_column_list=ignored_col_list,
                                            sampling=False,
                                            custom_algorithm=custom_algorithm,
                                            deployed_model_name=model_name)

try:
    rule_extractor_model = MltaskDeployment.deploy_mltask(
        deployed_model_name=model_name, ml_task=ml_task, project=project, train_dataset=trainset_name)
except:
    raise ValueError(
        'Training phase failed. Please go to the Lab section for more details.')


##########################################################################
# RETRIEVE THE TRAINED MODEL
##########################################################################

rules_df = rule_extractor_model.get_predictor().clf.retrieve_rules()

# Recipe outputs
output_ = get_output_names_for_role('Rules')
rules = dataiku.Dataset(output_[0])

if ml_type == 'clf':

    final_rule_df = rules_df.loc[
        :, ['rule_verbal', 'rule', 'support', 'precision']]
    final_rule_df['support'] = final_rule_df['support'] * 100
    final_rule_df['precision'] = final_rule_df['precision'] * 100
    final_rule_df = final_rule_df.rename(
        columns={'precision': 'label_1_percentage'})
    rules.write_with_schema(final_rule_df)

else:

    final_rule_df = rules_df.loc[
        :, ['rule_verbal', 'rule', 'support', 'target_mean', 'target_variance']]
    final_rule_df['support'] = final_rule_df['support'] * 100
    rules.write_with_schema(final_rule_df)
