# -*- coding: utf-8 -*-

import dataiku
from dataiku import pandasutils as pdu
import pandas as pd
import numpy as np
from sklearn.linear_model import Lasso as Lasso
from sklearn.linear_model import LogisticRegression as LogisticRegression
from sklearn.base import BaseEstimator
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO,  # avoid getting log from 3rd party module
                    format='RuleGeneration plugin %(levelname)s - %(message)s')


class RuleGeneratorRegularizer(BaseEstimator):

    def __init__(self, params=None):

        assert params is not None, "RuleGeneratorRegularizerParams instance is None."
        params.check()
        self.params = params

    def get_rules_list(self, df, feature_list, target):
        """
        Parameters
        ----------
        df : dataframe

        feature_list : list of string

        target : string

        Returns
        -------
        rules_list: list of string

        """
        if self.params.ml_type == 'clf':
            # if this is classification, we need to re-compute alpha to conform
            # the format of scikit-learn
            alpha = 1. / self.params.alpha
            clf = LogisticRegression(penalty='l1', C=alpha, random_state=1407)
            logger.info(
                "The rules will be regularized with Logistic Regression.")

        else:
            alpha = self.params.alpha
            clf = Lasso(alpha=alpha, random_state=1407)
            logger.info("The rules will be regularized with Lasso.")

        clf.fit(df[feature_list], df[target])

        if self.params.ml_type == 'clf':
            coef_ = list(clf.coef_[0])

        else:
            coef_ = clf.coef_

        rules_list = []  # list of rule columns

        for i in xrange(len(coef_)):
            if abs(coef_[i]) > 0:
                rules_list.append(feature_list[i])

        return rules_list

    def get_rules_df(self, rules_list, original_rules_df):
        """
        Filter rules in the original_rules_df dataframe based on those that are in rules_list.

        Parameters
        ----------
        rules_list : list of string
                list of tree rules

        original_rules_df: dataframe

        Returns
        -------
        filtered_ruled_df : dataframe
                The filtered rules dataframe.

        """

        filtered_ruled_df = original_rules_df.copy()

        # scan for rules that should be kept, give them flag of value 1
        if self.params.ml_type == 'clf':
            filtered_ruled_df['to_keep'] = filtered_ruled_df[['rule_verbal', 'support', 'precision', 'impurity']].apply(
                lambda x: 1 if x[0] in rules_list else 0, axis=1)
        else:
            filtered_ruled_df['to_keep'] = filtered_ruled_df[['rule_verbal',
                                                              'support',
                                                              'target_mean',
                                                              'target_variance']].apply(lambda x: 1 if x[0] in rules_list else 0,
                                                                                        axis=1)

        filtered_ruled_df = filtered_ruled_df[
            filtered_ruled_df['to_keep'] == 1]
        del filtered_ruled_df['to_keep']
        filtered_ruled_df = filtered_ruled_df.groupby(
            'rule_verbal').max().reset_index()

        return filtered_ruled_df
