# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
from sklearn.linear_model import Lasso as Lasso
from sklearn.linear_model import LogisticRegression as LogisticRegression
from sklearn.ensemble import RandomForestClassifier as RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor as RandomForestRegressor
from collections import Counter
from sklearn import tree
from scipy import sparse
from datetime import datetime
from sklearn.metrics import roc_auc_score as roc_auc_score
from sklearn.metrics import r2_score as r2_score
import time
import warnings
import os


def copytree(src, dst, symlinks=False, ignore=None):
    """
    Copy a folder. Thank you StackOverflow.

    """
    import shutil

    if os.path.isdir(dst):
        print('Delete and recreate the package inside dss lib folder')
        shutil.rmtree(dst)
        os.makedirs(dst)

    else:
        print('Copy package to dss lib folder')
        os.makedirs(dst)

    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def retrieve_features_from_rules(rules_df):
    """
    Retrieve features list from the rules dataframe.

    Parameters
    ----------
    rules_df : dataframe
        Dataframe of rules.


    Returns
    -------
    new_feature_list : list of string
        List of extracted features.

    """

    raw_feature_list = rules_df['rules_clean'].values.tolist()
    raw_feature_list = [x.replace('between', '') for x in raw_feature_list]
    raw_feature_list = [x.replace('above', '') for x in raw_feature_list]
    raw_feature_list = [x.replace('below', '') for x in raw_feature_list]
    raw_feature_list = [x.split(' & ') for x in raw_feature_list]

    selected_feature_list = []
    for x in raw_feature_list:  # retrieve final feature list

        l_ = [y.split(' ') for y in x]
        l_ = [y[0] for y in l_]
        l_ = list(set(l_))

        selected_feature_list = selected_feature_list + l_

    new_feature_list = list(set(selected_feature_list))

    return new_feature_list


def evaluateRuleClf(target_col, rule, X):
    """
    Help evaluate rule performance (precision and support) for classification.

    Parameters
    ----------

    target_col: string
        Name of the target column.

    rule : string
        A string that describes all the criterias to get to a specific node. 

    X : dataframe
        Dataframe with features and target.


    Returns
    -------

    support: float 
        Existence's ratio.

    precision: float
        Ratio of label 1 in the returned query.


    """

    val_count = X.query(rule)[target_col].value_counts()

    try:

        support = val_count.sum() / float(len(X))

        if len(val_count) > 1:
            precision = float(val_count[1]) / (val_count[0] + val_count[1])

        elif val_count.index[0]:  # if there is only label 1
            precision = 1

        else:  # if there is only label 0
            precision = 0

        return support, precision

    except IndexError:
        return np.nan, np.nan


def evaluateRuleReg(target_col, rule, X):
    """
    Help evaluate rule performance (variance and support) for regression.  


    Parameters
    ----------

    target_col: string
        Name of the target column.

    rule : string
        A string that describes all the criterias to get to a specific node. 


    X : dataframe
        Dataframe with features and target.


    Returns
    -------

    support: float
        Existence's ratio.

    target_mean: float
        Target's mean.

    target_variance: float
        Target's variance

    """

    val_count = X.query(rule)[target_col].value_counts()
    target_mean = X.query(rule)[target_col].mean()
    target_variance = X.query(rule)[target_col].var()

    try:
        if len(val_count) > 1:

            support = val_count.sum() / float(len(X))

        else:  # if there is only one (or zero) value ? wtf

            support = val_count.values[0] / float(len(X))
            target_variane = 0

        return support, target_mean, target_variance

    except IndexError:
        return np.nan, np.nan, np.nan
