# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
from sklearn.linear_model import Lasso as Lasso
from sklearn.linear_model import LogisticRegression as LogisticRegression
from sklearn.ensemble import RandomForestClassifier as RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor as RandomForestRegressor
from collections import Counter
from sklearn import tree
from scipy import sparse
from datetime import datetime
from sklearn.metrics import roc_auc_score as roc_auc_score
from sklearn.metrics import r2_score as r2_score
import time
import warnings

warnings.filterwarnings('ignore')


class treeExtract():

    """ 
    This class represents a Decision Tree estimator and encapsulate the path of rules to each of its nodes.

    """

    def __init__(self, clf, columns=None):
        """ 
        Attributes
        ----------

        clf : DecisionTree's instance
            The tree estimator.

        columns : dict
            Colunm's name in the input table, with their index

        extracted_tree: dataframe
            Dataframe of tree structure's attributes (left_child, right_child, threshold,...)
        """

        self.clf = clf
        self.columns = {}
        if type(columns) == list:
            for i, name in enumerate(columns):
                # for example {0: 'customer_id', 1: 'country', ...}
                self.columns[i] = name
        elif type(columns) == dict:
            self.columns = columns
        else:
            for i in xrange(1000):  # if no input, create 1000 columns (?)
                self.columns[i] = "col_" + str(i)

        self.extracted_tree = None

    def get_tree(self, ml_type):
        """
        The decision estimator has an attribute called tree_ which stores the entire
        tree structure and allows access to low level attributes. The binary tree
        tree_ is represented as a number of parallel arrays. The i-th element of each
        array holds information about the node `i`. Node 0 is the tree's root.

        Among those arrays, we have:
           - left_child: id of the left child of the node
           - right_child: id of the right child of the node
           - feature: feature used for splitting the node


        This function aims at retrieving that tree structure and storing some of its attributes 
        in the Dataframe called ``extracted_tree``.


        Parameters
        ----------

        ml_type : string
            Classification ('clf') or Regression ('reg')

        """

        t = self.clf.tree_  # retrieve the tree structure

        if ml_type == 'clf':
            cols = ['feature', 'children_left', 'children_right',
                    'threshold', 'impurity', 'n_node_samples']
            tree = np.vstack((t.feature, t.children_left, t.children_right,
                              t.threshold, t.impurity, t.n_node_samples))

        elif ml_type == 'reg':
            cols = ['feature', 'children_left',
                    'children_right', 'threshold', 'n_node_samples']
            tree = np.vstack((t.feature, t.children_left,
                              t.children_right, t.threshold, t.n_node_samples))

        # t.value contains the constant prediction value of each node. It seems
        # that we don't use this variable anyway.
        repartition = t.value.flatten()

        self.extracted_tree = pd.DataFrame(np.transpose(tree), columns=cols)

    def extract_tree(self):
        """
        Update the ``extracted_tree`` df so that for each node (row in df), we have all information
        concerning the path of node's id and of criteria (rule).
        """

        df_tree = self.extracted_tree
        clf = self.clf
        df_tree['estimator'] = id(clf)

        # map from feature's id to feature's name
        df_tree['feature'] = df_tree.feature.map(self.columns)

        df_tree['name_index'] = df_tree.apply(self.name_index, axis=1)

        df_tree['parent'] = [np.nan] + range(1, len(df_tree))
        dad_map_left = {int(v): str(int(k)) + 'l' for k,
                        v in df_tree.children_left.to_dict().items() if v > 0}
        dad_map_right = {int(v): str(int(k)) + 'r' for k,
                         v in df_tree.children_right.to_dict().items() if v > 0}

        dad_map_left.update(dad_map_right)
        # map the node index to the parent's node index
        df_tree['parent'] = df_tree.parent.map(dad_map_left)

        df_tree = df_tree.fillna(0)

        df_tree['depth'] = df_tree.parent.apply(lambda x: self.get_depth(x))
        df_tree.depth.ix[0] = 0

        # for each node (each row), "path" is the criteria in its parent node
        # to get to it
        df_tree['path'] = df_tree.apply(lambda x: self.get_path(x), axis=1)

        # now build the full path of rules
        df_tree['complete_path'] = df_tree.apply(
            lambda x: self.recurse_path(df_tree, x), axis=1)
        # still not sure why we need the following line ...
        df_tree['complete_path'] = df_tree[
            'complete_path'] + df_tree['feature'].astype(str)
        # reverse the elements in the rule so that they are in top down (root
        # down) order
        df_tree['complete_path'] = df_tree.complete_path.apply(self.clean_path)

        # do the same thing, but now we just build the path of node's id
        df_tree['complete_pathID'] = df_tree.apply(
            lambda x: self.recurse_pathID(df_tree, x), axis=1)
        df_tree['complete_pathID'] = df_tree[
            'complete_pathID'] + df_tree['parent'].astype(str)
        df_tree['complete_pathID'] = df_tree.complete_pathID.apply(
            self.clean_path)

        df_tree['last_feature'] = df_tree.complete_path.apply(
            self.get_last_feature)

        df_tree.parent.ix[0] = "root"
        mapper_rule = df_tree[['parent', 'path']].set_index('parent').to_dict()[
            "path"]
        df_tree = df_tree[1:]
        df_tree['rule'] = df_tree.complete_pathID.apply(
            lambda x: self.map_rules(x, mapper_rule))  # rule in query format
        df_tree["rule_verbal"] = df_tree.rule.apply(
            self.grammaire)  # rule in verbal format

        # not necessary as this was just a copy by reference.
        self.extracted_tree = df_tree

    def fit(self, ml_type):
        """
        - Step 1 : retrieve the tree's structure from the decision tree estimator.
        - Step 2: for each node in the tree, we compute its path of rules and previous nodes ids.


        Parameters
        ----------

        ml_type : string
            Classification ('clf') or Regression ('reg')

        """

        self.get_tree(ml_type=ml_type)
        self.extract_tree()

    def name_index(self, df):
        """
        Scan categorical column and store its values.

        In case the column has more distinct values than allowed, apply clipping method.

        Parameters
        ----------

        df : dataframe


        Returns
        -------



        """

        return str([df['children_left'], df['children_right']])

    def get_depth(self, z):
        """
        Get the depth of each node, based on their parent's id. 
        This is a recursive method that start in the bottom and then climb back up till the tree's root.


        Parameters
        ----------

        z : string
            Value of the column ``parent`` in the df_tree dataframe (the one that stocks information of one decision tree)

        Returns
        -------

        int value


        """

        # because value in "parent" column is in the format 1l, 2r ...
        if "r" in str(z) or "l" in str(z):
            z = int(z[:-1])
        if str(int(z)) == "0":  # tree's root
            return 1
        else:
            return 1 + self.get_depth(self.extracted_tree.parent.ix[int(z)][:-1])

    def get_path(self, z):
        """
        Get the split criteria 

        Parameters
        ----------

        z : pandas Series
            Each row (node) in the df_tree dataframe (the one that stocks information of one decision tree). 


        Returns
        -------

        a string, corresponding to the split criteria.


        """

        if z.ix['depth'] > 0:
            node_sign = z.ix['parent']  # exemple: 1l, 2r, ...
            node = node_sign[:-1]  # node's id
            sign = node_sign[-1:]  # left child or right child of its parent ?

            return self.get_insight(self.extracted_tree.ix[int(node)], sign)
        else:  # when do we have this case ?
            return ""

    def get_insight(self, node, sign):
        """
        Return a string that describes the split criteria of the actual node to the left child or right child.

        Parameters
        ----------

        node : pandas Series
            a row in the extracted_tree df.

        sign : string
            criteria for left child or right child ? 

        Returns
        -------



        """

        if sign == 'r':
            return str(node.ix['feature']) + ' > ' + str(round(node.ix['threshold'], 2))
        else:  # l
            return str(node.ix['feature']) + ' <= ' + str(round(node.ix['threshold'], 2))

    def recurse_path(self, df_tree, df):
        """
        Return the complete rules (path) for each node in a decision tree.


        Parameters
        ----------

        df_tree : the extracted decision tree from sklearn estimator. -> WHY DON'T JUST USE self.extracted_tree like in the get_depth function ? 

        df : pandas Series
            Each row (node) in the df_tree dataframe.


        Returns
        -------

        string, the full path


        """

        if df['path'] == "":  # root
            return ";"
        else:
            return df['path'] + "-" + self.recurse_path(df_tree, df_tree.ix[int(df['parent'][:-1])])

    def recurse_pathID(self, df_tree, df):
        """
        Return the complete IDs path for each node in a decision tree.

        Parameters
        ----------

        df_tree : the extracted decision tree from sklearn estimator. 

        df : pandas Series
            Each row (node) in the df_tree dataframe.


        Returns
        -------

        string, the full path


        """
        if df['path'] == "":
            return ";"
        else:
            return df['parent'].replace('-', 'l').replace('+', 'r') + "-" + \
                self.recurse_pathID(df_tree, df_tree.ix[int(
                    df['parent'][:-1])])  # not sure why we need the 2 replace functions ...

    def clean_path(self, df):
        """
        Clean the result returned by ``recurse_path`` (path of rules) or ``recurse_pathID`` (path of IDs) (technically, it is not the raw result,
        a small operation was applied on the raw output to add the feature's name, cf extract_tree).

        Basically we reverse the order in the rules so that it is top down (from the root).

        Parameters
        ----------

        df : string (yeah because life is not that simple...)
            The complete rules (path) for each node in a decision tree.


        Returns
        -------

        df : string
            The cleaned path.

        """

        df = df.split(';')[0]
        df = df.split('-')
        j = "-"
        df = j.join(df[::-1])[1:]

        return df

    def clean_count_class(self, df):
        """
        Scan categorical column and store its values.

        In case the column has more distinct values than allowed, apply clipping method.

        Parameters
        ----------

        df : 


        Returns
        -------



        """

        if df > N + 1 or df < 1:
            return 0
        else:
            return int(df)

    def get_count_classe(self, df, classe):
        """
        Scan categorical column and store its values.

        In case the column has more distinct values than allowed, apply clipping method.

        Parameters
        ----------

        df : 

        classe : 


        Returns
        -------



        """

        if df['children_left'] == -1:
            return df.ix[classe]
        else:
            return df.ix[classe] + self.get_count_classe(self.extracted_tree.ix[df['children_left']], classe) + \
                self.get_count_classe(self.extracted_tree.ix[
                                      df['children_right']], classe)

    def get_last_feature(self, df):
        """
        Get the name of the last feature's used for splitting of each node.

        Parameters
        ----------

        df : pandas Series
            Each row (node) in the df_tree dataframe.


        Returns
        -------

        string, the feature's name.


        """

        return df.split('-')[-1:][0].replace('>', '-').replace('<', '-').split('-')[0].rstrip()

    def map_rules(self, df, mapper_rule):
        """
        Parameters
        ----------

        df : string
            Path of IDs to a particular node.


        mapper_rule : dict
            A dict conaining the split criteria for each node.


        Returns
        -------

        string 

        """

        rule = df.split('-')  # list of IDs or criteria
        t = [mapper_rule[r] for r in rule]  # list of rules
        s = ' & '
        return s.join(t)

    def grammaire(self, rule):
        """
        Parameters
        ----------

        rule : string
            For example: "Embarked_value_S <= 0.5 & Pclass_value_3 > 0.5" 


        Returns
        -------

        string, verbolised rule.


        """

        listR = rule.split(" & ")
        d_inf = {}
        d_sup = {}

        c_inf = {}
        c_sup = {}

        for i in listR:

            # sorry future me, but Im too lazy to re-write these codes
            if 'dummy:' in i:
                if "<=" in i:
                    f, tr = i.split(" <= ")
                    f1, f2, f3 = f.split(':')
                    c_inf[f2] = f3

                elif ">" in i:
                    f, tr = i.split(" > ")
                    f1, f2, f3 = f.split(':')
                    c_sup[f2] = f3

            else:
                if "<=" in i:
                    f, tr = i.split(" <= ")
                    d_inf[f] = tr

                elif ">" in i:
                    f, tr = i.split(" > ")
                    d_sup[f] = tr

        s = ""

        for f in sorted(list(set(d_inf.keys() + d_sup.keys()))):

            if f in d_inf and f in d_sup:
                s += f + " between " + d_sup[f] + " and " + d_inf[f]
            elif f in d_inf:
                s += f + " below " + d_inf[f]
            elif f in d_sup:
                s += f + " above " + d_sup[f]
            s += " & "

        for f in sorted(list(set(c_inf.keys() + c_sup.keys()))):

            if f in c_inf:
                s += f + " is not " + c_inf[f]
            elif f in c_sup:
                s += f + " is " + c_sup[f]
            s += " & "

        return s[:-3]
